@@ -1,10 +1,32 @@
+# encoding: UTF-8
+
 class RubyLexer
+
+  # :stopdoc:
+  RUBY19 = "".respond_to? :encoding
+
+  IDENT_CHAR_RE = if RUBY19 then
+                    /[\w\u0080-\u{10ffff}]/u
+                  else
+                    /[\w\x80-\xFF]/n
+                  end
+
+  IDENT_RE = /^#{IDENT_CHAR_RE}+/o
+
   attr_accessor :command_start
   attr_accessor :cmdarg
   attr_accessor :cond
-  attr_accessor :nest
+  attr_accessor :tern # TODO: rename ternary damnit... wtf
+  attr_accessor :string_nest
 
-  ESC_RE = /\\([0-7]{1,3}|x[0-9a-fA-F]{1,2}|M-[^\\]|(C-|c)[^\\]|[^0-7xMCc])/
+  ESC_RE = /\\((?>[0-7]{1,3}|x[0-9a-fA-F]{1,2}|M-[^\\]|(C-|c)[^\\]|[^0-7xMCc]))/u
+  # :startdoc:
+
+  ##
+  # What version of ruby to parse. 18 and 19 are the only valid values
+  # currently supported.
+
+  attr_accessor :version
 
   # Additional context surrounding tokens that both the lexer and
   # grammar use.
@@ -28,14 +50,21 @@ class RubyLexer
   # What handles warnings
   attr_accessor :warnings
 
+  attr_accessor :space_seen
+  attr_accessor :paren_nest
+  attr_accessor :brace_nest
+  attr_accessor :lpar_beg
+
   EOF = :eof_haha!
 
   # ruby constants for strings (should this be moved somewhere else?)
+
+  # :stopdoc:
   STR_FUNC_BORING = 0x00
   STR_FUNC_ESCAPE = 0x01 # TODO: remove and replace with REGEXP
   STR_FUNC_EXPAND = 0x02
   STR_FUNC_REGEXP = 0x04
-  STR_FUNC_AWORDS = 0x08
+  STR_FUNC_QWORDS = 0x08
   STR_FUNC_SYMBOL = 0x10
   STR_FUNC_INDENT = 0x20 # <<-HEREDOC
 
@@ -49,6 +78,7 @@ class RubyLexer
   TOKENS = {
     "!"   => :tBANG,
     "!="  => :tNEQ,
+    # "!@"  => :tUBANG,
     "!~"  => :tNMATCH,
     ","   => :tCOMMA,
     ".."  => :tDOT2,
@@ -58,7 +88,9 @@ class RubyLexer
     "===" => :tEQQ,
     "=>"  => :tASSOC,
     "=~"  => :tMATCH,
+    "->"  => :tLAMBDA,
   }
+  # :startdoc:
 
   # How the parser advances to the next token.
   #
@@ -91,7 +123,7 @@ class RubyLexer
   end
 
   def fix_arg_lex_state
-    self.lex_state = if lex_state == :expr_fname || lex_state == :expr_dot
+    self.lex_state = if in_lex_state? :expr_fname, :expr_dot then
                        :expr_arg
                      else
                        :expr_beg
@@ -101,9 +133,9 @@ class RubyLexer
   def heredoc here # 63 lines
     _, eos, func, last_line = here
 
-    indent  = (func & STR_FUNC_INDENT) != 0
+    indent  = (func & STR_FUNC_INDENT) != 0 ? "[ \t]*" : nil
     expand  = (func & STR_FUNC_EXPAND) != 0
-    eos_re  = indent ? /[ \t]*#{eos}(\r?\n|\z)/ : /#{eos}(\r?\n|\z)/
+    eos_re  = /#{indent}#{Regexp.escape eos}(\r*\n|\z)/
     err_msg = "can't match #{eos_re.inspect} anywhere in "
 
     rb_compile_error err_msg if
@@ -130,7 +162,7 @@ class RubyLexer
         string_buffer << '#'
       end
 
-      until src.scan(eos_re) do
+      begin
         c = tokadd_string func, "\n", nil
 
         rb_compile_error err_msg if
@@ -145,10 +177,7 @@ class RubyLexer
 
         rb_compile_error err_msg if
           src.eos?
-      end
-
-      # tack on a NL after the heredoc token - FIX NL should not be needed
-      src.unread_many(eos + "\n") # TODO: remove this... stupid stupid stupid
+      end until src.check(eos_re)
     else
       until src.check(eos_re) do
         string_buffer << src.scan(/.*(\n|\z)/)
@@ -168,11 +197,9 @@ class RubyLexer
     self.string_buffer = []
 
     case
-    when src.scan(/(-?)(['"`])(.*?)\2/) then
+    when src.scan(/(-?)([\'\"\`])(.*?)\2/) then
       term = src[2]
-      unless src[1].empty? then
-        func |= STR_FUNC_INDENT
-      end
+      func |= STR_FUNC_INDENT unless src[1].empty?
       func |= case term
               when "\'" then
                 STR_SQUOTE
@@ -182,9 +209,9 @@ class RubyLexer
                 STR_XQUOTE
               end
       string_buffer << src[3]
-    when src.scan(/-?(['"`])(?!\1*\Z)/) then
+    when src.scan(/-?([\'\"\`])(?!\1*\Z)/) then
       rb_compile_error "unterminated here document identifier"
-    when src.scan(/(-?)(\w+)/) then
+    when src.scan(/(-?)(#{IDENT_CHAR_RE}+)/) then
       term = '"'
       func |= STR_DQUOTE
       unless src[1].empty? then
@@ -195,12 +222,10 @@ class RubyLexer
       return nil
     end
 
-    if src.check(/.*\n/) then
+    if src.scan(/.*\n/) then
       # TODO: think about storing off the char range instead
-      line = src.string[src.pos, src.matched_size]
-      src.string[src.pos, src.matched_size] = "\n"
+      line = src.matched
       src.extra_lines_added += 1
-      src.pos += 1
     else
       line = nil
     end
@@ -216,10 +241,20 @@ class RubyLexer
     end
   end
 
-  def initialize
-    self.cond = RubyParser::StackState.new(:cond)
-    self.cmdarg = RubyParser::StackState.new(:cmdarg)
-    self.nest = 0
+  def in_lex_state?(*states)
+    states.include? lex_state
+  end
+
+  def initialize v = 18
+    self.version = v
+    self.cond   = RubyParserStuff::StackState.new(:cond)
+    self.cmdarg = RubyParserStuff::StackState.new(:cmdarg)
+    self.tern   = RubyParserStuff::StackState.new(:tern)
+    self.string_nest = 0
+    self.paren_nest = 0
+    self.brace_nest = 0
+    self.lpar_beg = nil
+
     @comments = []
 
     reset
@@ -227,11 +262,13 @@ class RubyLexer
 
   def int_with_base base
     rb_compile_error "Invalid numeric format" if src.matched =~ /__/
+
     self.yacc_value = src.matched.to_i(base)
     return :tINTEGER
   end
 
   def lex_state= o
+    # warn "wtf lex_state = #{o.inspect} from #{caller.first}"
     raise "wtf\?" unless Symbol === o
     @lex_state = o
   end
@@ -251,14 +288,14 @@ class RubyLexer
     self.lex_state = :expr_end
 
     case
-    when src.scan(/[+-]?0[xbd]\b/) then
+    when src.scan(/[+-]?0[xXbBdD]\b/) then
       rb_compile_error "Invalid numeric format"
+    when src.scan(/[+-]?(?:(?:[1-9][\d_]*|0)(?!\.\d)\b|0[Dd][0-9_]+)/) then
+      int_with_base(10)
     when src.scan(/[+-]?0x[a-f0-9_]+/i) then
       int_with_base(16)
-    when src.scan(/[+-]?0b[01_]+/) then
+    when src.scan(/[+-]?0[Bb][01_]+/) then
       int_with_base(2)
-    when src.scan(/[+-]?0d[0-9_]+/) then
-      int_with_base(10)
     when src.scan(/[+-]?0[Oo]?[0-7_]*[89]/) then
       rb_compile_error "Illegal octal digit."
     when src.scan(/[+-]?0[Oo]?[0-7_]+|0[Oo]/) then
@@ -272,9 +309,7 @@ class RubyLexer
       end
       self.yacc_value = number.to_f
       :tFLOAT
-    when src.scan(/[+-]?0\b/) then
-      int_with_base(10)
-    when src.scan(/[+-]?[\d_]+\b/) then
+    when src.scan(/[+-]?[0-9_]+(?![e])/) then
       int_with_base(10)
     else
       rb_compile_error "Bad number format"
@@ -309,10 +344,10 @@ class RubyLexer
                                 [:tSTRING_BEG,   STR_SQUOTE]
                               when 'W' then
                                 src.scan(/\s*/)
-                                [:tWORDS_BEG,    STR_DQUOTE | STR_FUNC_AWORDS]
+                                [:tWORDS_BEG,    STR_DQUOTE | STR_FUNC_QWORDS]
                               when 'w' then
                                 src.scan(/\s*/)
-                                [:tAWORDS_BEG,   STR_SQUOTE | STR_FUNC_AWORDS]
+                                [:tQWORDS_BEG,   STR_SQUOTE | STR_FUNC_QWORDS]
                               when 'x' then
                                 [:tXSTRING_BEG,  STR_XQUOTE]
                               when 'r' then
@@ -320,9 +355,13 @@ class RubyLexer
                               when 's' then
                                 self.lex_state  = :expr_fname
                                 [:tSYMBEG,       STR_SSYM]
+                              when 'I' then
+                                [:tSYMBOLS_BEG, STR_DQUOTE | STR_FUNC_QWORDS]
+                              when 'i' then
+                                [:tQSYMBOLS_BEG, STR_SQUOTE | STR_FUNC_QWORDS]
                               end
 
-    rb_compile_error "Bad %string type. Expected [Qqwxr\W], found '#{c}'." if
+    rb_compile_error "Bad %string type. Expected [Qq\Wwxrs], found '#{c}'." if
       token_type.nil?
 
     self.lex_strterm = [:strterm, string_type, nnd, beg]
@@ -336,9 +375,9 @@ class RubyLexer
     space = false # FIX: remove these
     func = string_type
     paren = open
-    term_re = Regexp.escape term
+    term_re = @@regexp_cache[term]
 
-    awords = (func & STR_FUNC_AWORDS) != 0
+    qwords = (func & STR_FUNC_QWORDS) != 0
     regexp = (func & STR_FUNC_REGEXP) != 0
     expand = (func & STR_FUNC_EXPAND) != 0
 
@@ -347,11 +386,11 @@ class RubyLexer
       return :tSTRING_END
     end
 
-    space = true if awords and src.scan(/\s+/)
+    space = true if qwords and src.scan(/\s+/)
 
-    if self.nest == 0 && src.scan(/#{term_re}/) then
-      if awords then
-        quote[1] = nil
+    if self.string_nest == 0 && src.scan(/#{term_re}/) then
+      if qwords then
+        quote[1] = nil # TODO: make struct
         return :tSPACE
       elsif regexp then
         self.yacc_value = self.regx_options
@@ -392,7 +431,7 @@ class RubyLexer
 
   def rb_compile_error msg
     msg += ". near line #{self.lineno}: #{src.rest[/^.*/].inspect}"
-    raise SyntaxError, msg
+    raise RubyParser::SyntaxError, msg
   end
 
   def read_escape # 51 lines
@@ -418,7 +457,7 @@ class RubyLexer
     when src.scan(/s/) then                   # space
       " "
     when src.scan(/[0-7]{1,3}/) then          # octal constant
-      src.matched.to_i(8).chr
+      (src.matched.to_i(8) & 0xFF).chr
     when src.scan(/x([0-9a-fA-F]{1,2})/) then # hex constant
       src[1].to_i(16).chr
     when src.check(/M-\\[\\MCc]/) then
@@ -441,6 +480,8 @@ class RubyLexer
       c = src[2]
       c[0] = (c[0].ord & 0x9f).chr
       c
+    when src.scan(/^[89]/i) then # bad octal or hex... MRI ignores them :(
+      src.matched
     when src.scan(/[McCx0-9]/) || src.eos? then
       rb_compile_error("Invalid escape character syntax")
     else
@@ -473,6 +514,14 @@ class RubyLexer
     @lex_state = nil
   end
 
+  def ruby18
+    Ruby18Parser === parser
+  end
+
+  def ruby19
+    Ruby19Parser === parser
+  end
+
   def src= src
     raise "bad src: #{src.inspect}" unless String === src
     @src = RPStringScanner.new(src)
@@ -498,41 +547,47 @@ class RubyLexer
     end
   end
 
+  @@regexp_cache = Hash.new { |h,k| h[k] = Regexp.new(Regexp.escape(k)) }
+  @@regexp_cache[nil] = nil
+
   def tokadd_string(func, term, paren) # 105 lines
-    awords = (func & STR_FUNC_AWORDS) != 0
+    qwords = (func & STR_FUNC_QWORDS) != 0
     escape = (func & STR_FUNC_ESCAPE) != 0
     expand = (func & STR_FUNC_EXPAND) != 0
     regexp = (func & STR_FUNC_REGEXP) != 0
     symbol = (func & STR_FUNC_SYMBOL) != 0
 
-    paren_re = paren.nil? ? nil : Regexp.new(Regexp.escape(paren))
-    term_re  = Regexp.new(Regexp.escape(term))
+    paren_re = @@regexp_cache[paren]
+    term_re  = @@regexp_cache[term]
 
     until src.eos? do
       c = nil
       handled = true
+
       case
-      when self.nest == 0 && src.scan(term_re) then
-        src.pos -= 1
-        break
       when paren_re && src.scan(paren_re) then
-        self.nest += 1
+        self.string_nest += 1
       when src.scan(term_re) then
-        self.nest -= 1
-      when awords && src.scan(/\s/) then
+        if self.string_nest == 0 then
+          src.pos -= 1
+          break
+        else
+          self.string_nest -= 1
+        end
+      when expand && src.scan(/#(?=[\$\@\{])/) then
         src.pos -= 1
         break
-      when expand && src.scan(/#(?=[\$\@\{])/) then
+      when qwords && src.scan(/\s/) then
         src.pos -= 1
         break
       when expand && src.scan(/#(?!\n)/) then
         # do nothing
       when src.check(/\\/) then
         case
-        when awords && src.scan(/\\\n/) then
+        when qwords && src.scan(/\\\n/) then
           string_buffer << "\n"
           next
-        when awords && src.scan(/\\\s/) then
+        when qwords && src.scan(/\\\s/) then
           c = ' '
         when expand && src.scan(/\\\n/) then
           next
@@ -552,17 +607,20 @@ class RubyLexer
           end
         else
           handled = false
-        end
+        end # inner /\\/ case
       else
         handled = false
-      end # case
+      end # top case
 
       unless handled then
-
         t = Regexp.escape term
         x = Regexp.escape(paren) if paren && paren != "\000"
-        re = if awords then
-               /[^#{t}#{x}\#\0\\\n\ ]+|./ # |. to pick up whatever
+        re = if qwords then
+               if RUBY19 then
+                 /[^#{t}#{x}\#\0\\\s]+|./ # |. to pick up whatever
+               else
+                 /[^#{t}#{x}\#\0\\\s\v]+|./ # argh. 1.8's \s doesn't pick up \v
+               end
              else
                /[^#{t}#{x}\#\0\\]+|./
              end
@@ -580,44 +638,48 @@ class RubyLexer
     c ||= src.matched
     c = RubyLexer::EOF if src.eos?
 
-
     return c
   end
 
-  def unescape s
+  ESCAPES = {
+    "a"    => "\007",
+    "b"    => "\010",
+    "e"    => "\033",
+    "f"    => "\f",
+    "n"    => "\n",
+    "r"    => "\r",
+    "s"    => " ",
+    "t"    => "\t",
+    "v"    => "\13",
+    "\\"   => '\\',
+    "\n"   => "",
+    "C-\?" => 127.chr,
+    "c\?"  => 127.chr,
+  }
 
-    r = {
-      "a"    => "\007",
-      "b"    => "\010",
-      "e"    => "\033",
-      "f"    => "\f",
-      "n"    => "\n",
-      "r"    => "\r",
-      "s"    => " ",
-      "t"    => "\t",
-      "v"    => "\13",
-      "\\"   => '\\',
-      "\n"   => "",
-      "C-\?" => 127.chr,
-      "c\?"  => 127.chr,
-    }[s]
+  def unescape s
+    r = ESCAPES[s]
 
     return r if r
 
-    case s
-    when /^[0-7]{1,3}/ then
-      $&.to_i(8).chr
-    when /^x([0-9a-fA-F]{1,2})/ then
-      $1.to_i(16).chr
-    when /^M-(.)/ then
-      ($1[0].ord | 0x80).chr
-    when /^(C-|c)(.)/ then
-      ($2[0].ord & 0x9f).chr
-    when /^[McCx0-9]/ then
-      rb_compile_error("Invalid escape character syntax")
-    else
-      s
-    end
+    x = case s
+        when /^[0-7]{1,3}/ then
+          ($&.to_i(8) & 0xFF).chr
+        when /^x([0-9a-fA-F]{1,2})/ then
+          $1.to_i(16).chr
+        when /^M-(.)/ then
+          ($1[0].ord | 0x80).chr
+        when /^(C-|c)(.)/ then
+          ($2[0].ord & 0x9f).chr
+        when /^[89a-f]/i then # bad octal or hex... ignore? that's what MRI does :(
+          s
+        when /^[McCx0-9]/ then
+          rb_compile_error("Invalid escape character syntax")
+        else
+          s
+        end
+    x.force_encoding "UTF-8" if RUBY19
+    x
   end
 
   def warning s
@@ -630,9 +692,8 @@ class RubyLexer
   # @return Description of the Returned Value
 
   def yylex # 826 lines
-
     c = ''
-    space_seen = false
+    self.space_seen = false
     command_state = false
     src = self.src
 
@@ -648,7 +709,7 @@ class RubyLexer
 
     loop do # START OF CASE
       if src.scan(/[\ \t\r\f\v]/) then # \s - \n + \v
-        space_seen = true
+        self.space_seen = true
         next
       elsif src.check(/[^a-zA-Z]/) then
         if src.scan(/\n|#/) then
@@ -661,26 +722,42 @@ class RubyLexer
               @comments << src.matched.gsub(/^ +#/, '#').gsub(/^ +$/, '')
             end
 
-            if src.eos? then
-              return RubyLexer::EOF
-            end
+            return RubyLexer::EOF if src.eos?
           end
 
           # Replace a string of newlines with a single one
           src.scan(/\n+/)
 
-          if [:expr_beg, :expr_fname,
-              :expr_dot, :expr_class].include? lex_state then
-            next
+          next if in_lex_state?(:expr_beg, :expr_value, :expr_class,
+                                :expr_fname, :expr_dot)
+
+          if src.scan(/([\ \t\r\f\v]*)\./) then
+            self.space_seen = true unless src[1].empty?
+
+            src.pos -= 1
+            next unless src.check(/\.\./)
           end
 
           self.command_start = true
           self.lex_state = :expr_beg
           return :tNL
         elsif src.scan(/[\]\)\}]/) then
+          if src.matched == "}" then
+            self.brace_nest -= 1
+          else
+            self.paren_nest -= 1
+          end
+
           cond.lexpop
           cmdarg.lexpop
-          self.lex_state = :expr_end
+          tern.lexpop
+
+          self.lex_state = if src.matched == ")" then
+                             :expr_endfn
+                           else
+                             :expr_endarg
+                           end
+
           self.yacc_value = src.matched
           result = {
             ")" => :tRPAREN,
@@ -688,6 +765,25 @@ class RubyLexer
             "}" => :tRCURLY
           }[src.matched]
           return result
+        elsif src.scan(/\!/) then
+          if in_lex_state?(:expr_fname, :expr_dot) then
+            self.lex_state = :expr_arg
+
+            if src.scan(/@/) then
+              self.yacc_value = "!@"
+              return :tUBANG
+            end
+          else
+            self.lex_state = :expr_beg
+          end
+
+          if src.scan(/[=~]/) then
+            self.yacc_value = "!#{src.matched}"
+          else
+            self.yacc_value = "!"
+          end
+
+          return TOKENS[self.yacc_value]
         elsif src.scan(/\.\.\.?|,|![=~]?/) then
           self.lex_state = :expr_beg
           tok = self.yacc_value = src.matched
@@ -701,19 +797,13 @@ class RubyLexer
             return :tDOT
           end
         elsif src.scan(/\(/) then
-          result = :tLPAREN2
-          self.command_start = true
+          result = if ruby18 then
+                     yylex_paren18
+                   else
+                     yylex_paren19
+                   end
 
-          if lex_state == :expr_beg || lex_state == :expr_mid then
-            result = :tLPAREN
-          elsif space_seen then
-            if lex_state == :expr_cmdarg then
-              result = :tLPAREN_ARG
-            elsif lex_state == :expr_arg then
-              warning("don't put space before argument parentheses")
-              result = :tLPAREN2
-            end
-          end
+          self.paren_nest += 1
 
           self.expr_beg_push "("
 
@@ -724,7 +814,6 @@ class RubyLexer
             tok = self.yacc_value = src.matched
             return TOKENS[tok]
           elsif src.scan(/\=begin(?=\s)/) then
-            # @comments << '=' << src.matched
             @comments << src.matched
 
             unless src.scan(/.*?\n=end( |\t|\f)*[^\n]*(\n|\z)/m) then
@@ -746,7 +835,7 @@ class RubyLexer
           self.lex_strterm = [:strterm, STR_DQUOTE, '"', "\0"] # TODO: question this
           self.yacc_value = "\""
           return :tSTRING_BEG
-        elsif src.scan(/\@\@?\w*/) then
+        elsif src.scan(/\@\@?#{IDENT_CHAR_RE}+/o) then
           self.token = src.matched
 
           rb_compile_error "`#{token}` is not allowed as a variable name" if
@@ -754,10 +843,7 @@ class RubyLexer
 
           return process_token(command_state)
         elsif src.scan(/\:\:/) then
-          if (lex_state == :expr_beg ||
-              lex_state == :expr_mid ||
-              lex_state == :expr_class ||
-              (lex_state.is_argument && space_seen)) then
+          if is_beg? || in_lex_state?(:expr_class) || is_space_arg? then
             self.lex_state = :expr_beg
             self.yacc_value = "::"
             return :tCOLON3
@@ -766,15 +852,16 @@ class RubyLexer
           self.lex_state = :expr_dot
           self.yacc_value = "::"
           return :tCOLON2
-        elsif lex_state != :expr_end && lex_state != :expr_endarg && src.scan(/:([a-zA-Z_]\w*(?:[?!]|=(?!>))?)/) then
+        elsif ! is_end? && src.scan(/:([a-zA-Z_]#{IDENT_CHAR_RE}*(?:[?!]|=(?==>)|=(?![=>]))?)/) then
+          # scanning shortcut to symbols
           self.yacc_value = src[1]
           self.lex_state = :expr_end
           return :tSYMBOL
         elsif src.scan(/\:/) then
           # ?: / then / when
-          if (lex_state == :expr_end || lex_state == :expr_endarg||
-              src.check(/\s/)) then
+          if is_end? || src.check(/\s/) then
             self.lex_state = :expr_beg
+            # TODO warn_balanced(":", "symbol literal");
             self.yacc_value = ":"
             return :tCOLON
           end
@@ -792,31 +879,39 @@ class RubyLexer
         elsif src.check(/[0-9]/) then
           return parse_number
         elsif src.scan(/\[/) then
+          self.paren_nest += 1
+
           result = src.matched
 
-          if lex_state == :expr_fname || lex_state == :expr_dot then
+          if in_lex_state? :expr_fname, :expr_dot then
             self.lex_state = :expr_arg
             case
             when src.scan(/\]\=/) then
+              self.paren_nest -= 1 # HACK? I dunno, or bug in MRI
               self.yacc_value = "[]="
               return :tASET
             when src.scan(/\]/) then
+              self.paren_nest -= 1 # HACK? I dunno, or bug in MRI
               self.yacc_value = "[]"
               return :tAREF
             else
               rb_compile_error "unexpected '['"
             end
-          elsif lex_state == :expr_beg || lex_state == :expr_mid then
+          elsif is_beg? then
+            self.tern.push false
             result = :tLBRACK
-          elsif lex_state.is_argument && space_seen then
+          elsif is_arg? && space_seen then
+            self.tern.push false
             result = :tLBRACK
+          else
+            result = :tLBRACK2
           end
 
           self.expr_beg_push "["
 
           return result
         elsif src.scan(/\'(\\.|[^\'])*\'/) then
-          self.yacc_value = src.matched[1..-2].gsub(/\\\\/, "\\").gsub(/\\'/, "'")
+          self.yacc_value = src.matched[1..-2].gsub(/\\\\/, "\\").gsub(/\\'/, "'") # "
           self.lex_state = :expr_end
           return :tSTRING
         elsif src.check(/\|/) then
@@ -838,11 +933,22 @@ class RubyLexer
             return :tPIPE
           end
         elsif src.scan(/\{/) then
-          result = if lex_state.is_argument || lex_state == :expr_end then
+          self.brace_nest += 1
+          if lpar_beg && lpar_beg == paren_nest then
+            self.lpar_beg = nil
+            self.paren_nest -= 1
+
+            expr_beg_push "{"
+
+            return :tLAMBEG
+          end
+
+          result = if is_arg? || in_lex_state?(:expr_end, :expr_endfn) then
                      :tLCURLY      #  block (primary)
-                   elsif lex_state == :expr_endarg then
+                   elsif in_lex_state?(:expr_endarg) then
                      :tLBRACE_ARG  #  block (expr)
                    else
+                     self.tern.push false
                      :tLBRACE      #  hash
                    end
 
@@ -850,6 +956,9 @@ class RubyLexer
           self.command_start = true unless result == :tLBRACE
 
           return result
+        elsif src.scan(/->/) then
+          self.lex_state = :expr_endfn
+          return :tLAMBDA
         elsif src.scan(/[+-]/) then
           sign = src.matched
           utype, type = if sign == "+" then
@@ -858,7 +967,7 @@ class RubyLexer
                           [:tUMINUS, :tMINUS]
                         end
 
-          if lex_state == :expr_fname || lex_state == :expr_dot then
+          if in_lex_state? :expr_fname, :expr_dot then
             self.lex_state = :expr_arg
             if src.scan(/@/) then
               self.yacc_value = "#{sign}@"
@@ -875,9 +984,8 @@ class RubyLexer
             return :tOP_ASGN
           end
 
-          if (lex_state == :expr_beg || lex_state == :expr_mid ||
-              (lex_state.is_argument && space_seen && !src.check(/\s/))) then
-            if lex_state.is_argument then
+          if (is_beg? || (is_arg? && space_seen && !src.check(/\s/))) then
+            if is_arg? then
               arg_ambiguous
             end
 
@@ -904,25 +1012,35 @@ class RubyLexer
             self.yacc_value = "**"
             return :tOP_ASGN
           elsif src.scan(/\*\*/) then
+            result = if is_space_arg? src.check(/./m) then
+                       warning "`**' interpreted as argument prefix"
+                       :tDSTAR
+                     elsif is_beg? then
+                       :tDSTAR
+                     else
+                       # TODO: warn_balanced("**", "argument prefix");
+                       :tPOW
+                     end
             self.yacc_value = "**"
             self.fix_arg_lex_state
-            return :tPOW
+            return result
           elsif src.scan(/\*\=/) then
             self.lex_state = :expr_beg
             self.yacc_value = "*"
             return :tOP_ASGN
           elsif src.scan(/\*/) then
-            result = if lex_state.is_argument && space_seen && src.check(/\S/) then
+            result = if is_space_arg? src.check(/./m) then
                        warning("`*' interpreted as argument prefix")
                        :tSTAR
-                     elsif lex_state == :expr_beg || lex_state == :expr_mid then
+                     elsif is_beg? then
                        :tSTAR
                      else
-                       :tSTAR2
+                       # TODO: warn_balanced("*", "argument prefix");
+                       :tSTAR2 # TODO: rename
                      end
+
             self.yacc_value = "*"
             self.fix_arg_lex_state
-
             return result
           end
         elsif src.check(/\</) then
@@ -940,13 +1058,11 @@ class RubyLexer
             self.yacc_value = "\<\<"
             return :tOP_ASGN
           elsif src.scan(/\<\</) then
-            if (! [:expr_end,    :expr_dot,
-                   :expr_endarg, :expr_class].include?(lex_state) &&
-                (!lex_state.is_argument || space_seen)) then
+            if (!in_lex_state?(:expr_dot, :expr_class) &&
+                !is_end? &&
+                (!is_arg? || space_seen)) then
               tok = self.heredoc_identifier
-              if tok then
-                return tok
-              end
+              return tok if tok
             end
 
             self.fix_arg_lex_state
@@ -993,8 +1109,10 @@ class RubyLexer
           self.lex_strterm = [:strterm, STR_XQUOTE, '`', "\0"]
           return :tXSTRING_BEG
         elsif src.scan(/\?/) then
-          if lex_state == :expr_end || lex_state == :expr_endarg then
-            self.lex_state = :expr_beg
+
+          if is_end? then
+            self.lex_state = ruby18 ? :expr_beg : :expr_value # HACK?
+            self.tern.push true
             self.yacc_value = "?"
             return :tEH
           end
@@ -1004,7 +1122,7 @@ class RubyLexer
           end
 
           if src.check(/\s|\v/) then
-            unless lex_state.is_argument then
+            unless is_arg? then
               c2 = { " " => 's',
                     "\n" => 'n',
                     "\t" => 't',
@@ -1018,11 +1136,13 @@ class RubyLexer
             end
 
             # ternary
-            self.lex_state = :expr_beg
+            self.lex_state = ruby18 ? :expr_beg : :expr_value # HACK?
+            self.tern.push true
             self.yacc_value = "?"
             return :tEH
           elsif src.check(/\w(?=\w)/) then # ternary, also
             self.lex_state = :expr_beg
+            self.tern.push true
             self.yacc_value = "?"
             return :tEH
           end
@@ -1033,8 +1153,14 @@ class RubyLexer
                 src.getch
               end
           self.lex_state = :expr_end
-          self.yacc_value = c[0].ord & 0xff
-          return :tINTEGER
+
+          if version == 18 then
+            self.yacc_value = c[0].ord & 0xff
+            return :tINTEGER
+          else
+            self.yacc_value = c
+            return :tSTRING
+          end
         elsif src.check(/\&/) then
           if src.scan(/\&\&\=/) then
             self.yacc_value = "&&"
@@ -1049,11 +1175,11 @@ class RubyLexer
             self.lex_state = :expr_beg
             return :tOP_ASGN
           elsif src.scan(/&/) then
-            result = if lex_state.is_argument && space_seen &&
+            result = if is_arg? && space_seen &&
                          !src.check(/\s/) then
                        warning("`&' interpreted as argument prefix")
                        :tAMPER
-                     elsif lex_state == :expr_beg || lex_state == :expr_mid then
+                     elsif in_lex_state? :expr_beg, :expr_mid then
                        :tAMPER
                      else
                        :tAMPER2
@@ -1064,7 +1190,7 @@ class RubyLexer
             return result
           end
         elsif src.scan(/\//) then
-          if lex_state == :expr_beg || lex_state == :expr_mid then
+          if is_beg? then
             self.lex_strterm = [:strterm, STR_REGEXP, '/', "\0"]
             self.yacc_value = "/"
             return :tREGEXP_BEG
@@ -1076,7 +1202,7 @@ class RubyLexer
             return :tOP_ASGN
           end
 
-          if lex_state.is_argument && space_seen then
+          if is_arg? && space_seen then
             unless src.scan(/\s/) then
               arg_ambiguous
               self.lex_strterm = [:strterm, STR_REGEXP, '/', "\0"]
@@ -1103,7 +1229,7 @@ class RubyLexer
           self.yacc_value = ";"
           return :tSEMI
         elsif src.scan(/\~/) then
-          if lex_state == :expr_fname || lex_state == :expr_dot then
+          if in_lex_state? :expr_fname, :expr_dot then
             src.scan(/@/)
           end
 
@@ -1112,14 +1238,14 @@ class RubyLexer
 
           return :tTILDE
         elsif src.scan(/\\/) then
-          if src.scan(/\n/) then
+          if src.scan(/\r?\n/) then
             self.lineno = nil
-            space_seen = true
+            self.space_seen = true
             next
           end
           rb_compile_error "bare backslash only allowed before newline"
         elsif src.scan(/\%/) then
-          if lex_state == :expr_beg || lex_state == :expr_mid then
+          if is_beg? then
             return parse_quote
           end
 
@@ -1129,9 +1255,7 @@ class RubyLexer
             return :tOP_ASGN
           end
 
-          if lex_state.is_argument && space_seen && ! src.check(/\s/) then
-            return parse_quote
-          end
+          return parse_quote if is_arg? && space_seen && ! src.check(/\s/)
 
           self.fix_arg_lex_state
           self.yacc_value = "%"
@@ -1184,7 +1308,7 @@ class RubyLexer
             return process_token(command_state)
           end
         elsif src.check(/\_/) then
-          if src.beginning_of_line? && src.scan(/\__END__(\n|\Z)/) then
+          if src.beginning_of_line? && src.scan(/\__END__(\r?\n|\Z)/) then
             self.lineno = nil
             return RubyLexer::EOF
           elsif src.scan(/\_\w*/) then
@@ -1197,25 +1321,74 @@ class RubyLexer
       if src.scan(/\004|\032|\000/) || src.eos? then # ^D, ^Z, EOF
         return RubyLexer::EOF
       else # alpha check
-        if src.scan(/\W/) then
-          rb_compile_error "Invalid char #{src.matched.inspect} in expression"
-        end
+        rb_compile_error "Invalid char #{src.rest[0].chr} in expression" unless
+          src.check IDENT_RE
       end
 
-      self.token = src.matched if self.src.scan(/\w+/)
+      self.token = src.matched if self.src.scan IDENT_RE
 
       return process_token(command_state)
     end
   end
 
-  def process_token(command_state)
+  def yylex_paren18
+    self.command_start = true
+    result = :tLPAREN2
 
-    token << src.matched if token =~ /^\w/ && src.scan(/[\!\?](?!=)/)
+    if in_lex_state? :expr_beg, :expr_mid then
+      result = :tLPAREN
+    elsif space_seen then
+      if in_lex_state? :expr_cmdarg then
+        result = :tLPAREN_ARG
+      elsif in_lex_state? :expr_arg then
+        self.tern.push false
+        warning "don't put space before argument parentheses"
+      end
+    else
+      self.tern.push false
+    end
+
+    result
+  end
+
+  def yylex_paren19
+    if is_beg? then
+      :tLPAREN
+    elsif is_space_arg? then
+      :tLPAREN_ARG
+    else
+      :tLPAREN2 # plain '(' in parse.y
+    end
+  end
+
+  def is_arg?
+    in_lex_state? :expr_arg, :expr_cmdarg
+  end
+
+  def is_end?
+    in_lex_state? :expr_end, :expr_endarg, :expr_endfn
+  end
+
+  def is_beg?
+    in_lex_state? :expr_beg, :expr_value, :expr_mid, :expr_class
+  end
+
+  # TODO #define IS_AFTER_OPERATOR() IS_lex_state(EXPR_FNAME | EXPR_DOT)
+
+  def is_space_arg? c = "x"
+    is_arg? and space_seen and c !~ /\s/
+  end
+
+  def is_label_possible? command_state
+    (in_lex_state?(:expr_beg) && !command_state) || is_arg?
+  end
+
+  def process_token(command_state)
+    token << src.matched if token =~ IDENT_RE && src.scan(/[\!\?](?!=)/)
 
     result = nil
     last_state = lex_state
 
-
     case token
     when /^\$/ then
       self.lex_state, result = :expr_end, :tGVAR
@@ -1227,7 +1400,7 @@ class RubyLexer
       if token =~ /[!?]$/ then
         result = :tFID
       else
-        if lex_state == :expr_fname then
+        if in_lex_state? :expr_fname then
           # ident=, not =~ => == or followed by =>
           # TODO test lexing of a=>b vs a==>b
           if src.scan(/=(?:(?![~>=])|(?==>))/) then
@@ -1243,9 +1416,27 @@ class RubyLexer
                    end
       end
 
-      unless lex_state == :expr_dot then
+      unless ruby18
+        if is_label_possible? command_state then
+          colon = src.scan(/:/)
+
+          if colon && src.peek(1) != ":" then
+            self.lex_state = :expr_beg
+            self.yacc_value = [token, src.lineno]
+            return :tLABEL
+          end
+
+          src.unscan if colon
+        end
+      end
+
+      unless in_lex_state? :expr_dot then
         # See if it is a reserved word.
-        keyword = RubyParser::Keyword.keyword token
+        keyword = if ruby18 then # REFACTOR need 18/19 lexer subclasses
+                    RubyParserStuff::Keyword.keyword18 token
+                  else
+                    RubyParserStuff::Keyword.keyword19 token
+                  end
 
         if keyword then
           state           = lex_state
@@ -1257,15 +1448,23 @@ class RubyLexer
             return keyword.id0
           end
 
+          self.command_start = true if lex_state == :expr_beg
+
           if keyword.id0 == :kDO then
-            self.command_start = true
+            if lpar_beg && lpar_beg == paren_nest then
+              self.lpar_beg = nil
+              self.paren_nest -= 1
+
+              return :kDO_LAMBDA
+            end
+
             return :kDO_COND  if cond.is_in_state
             return :kDO_BLOCK if cmdarg.is_in_state && state != :expr_cmdarg
-            return :kDO_BLOCK if state == :expr_endarg
+            return :kDO_BLOCK if [:expr_beg, :expr_endarg].include? state
             return :kDO
           end
 
-          return keyword.id0 if state == :expr_beg or state == :expr_value
+          return keyword.id0 if [:expr_beg, :expr_value].include? state
 
           self.lex_state = :expr_beg if keyword.id0 != keyword.id1
 
@@ -1273,24 +1472,30 @@ class RubyLexer
         end
       end
 
-      if (lex_state == :expr_beg || lex_state == :expr_mid ||
-          lex_state == :expr_dot || lex_state == :expr_arg ||
-          lex_state == :expr_cmdarg) then
-        if command_state then
-          self.lex_state = :expr_cmdarg
+      # TODO:
+      # if (mb == ENC_CODERANGE_7BIT && lex_state != EXPR_DOT) {
+
+      self.lex_state =
+        if is_beg? || is_arg? || in_lex_state?(:expr_dot) then
+          if command_state then
+            :expr_cmdarg
+          else
+            :expr_arg
+          end
+        elsif !ruby18 && in_lex_state?(:expr_fname) then
+          :expr_endfn
         else
-          self.lex_state = :expr_arg
+          :expr_end
         end
-      else
-        self.lex_state = :expr_end
-      end
+
     end
 
     self.yacc_value = token
 
-
-    self.lex_state = :expr_end if
-      last_state != :expr_dot && self.parser.env[token.to_sym] == :lvar
+    if (![:expr_dot, :expr_fname].include?(last_state) &&
+        self.parser.env[token.to_sym] == :lvar) then
+      self.lex_state = :expr_end
+    end
 
     return result
   end
