@@ -1,24 +1,50 @@
 #!/usr/local/bin/ruby
+# encoding: US-ASCII
 
 require 'rubygems'
 require 'minitest/autorun'
 require 'ruby_lexer'
-require 'ruby_parser'
+require 'ruby18_parser'
+require 'ruby20_parser'
 
-class TestRubyLexer < MiniTest::Unit::TestCase
-  alias :deny :refute
+class TestRubyLexer < Minitest::Test
+  attr_accessor :processor, :lex, :parser_class
+
+  alias :lexer :lex # lets me copy/paste code from parser
+  alias :lexer= :lex=
 
   def setup
-    p = RubyParser.new
-    @lex = p.lexer
-    @lex.src = "blah blah"
-    @lex.lex_state = :expr_beg
+    setup_lexer_class Ruby20Parser
+  end
+
+  def setup_lexer input, exp_sexp = nil
+    setup_new_parser
+    lex.src = input
+    lex.lex_state = :expr_beg
+    assert_equal exp_sexp, processor.class.new.parse(input) if exp_sexp
+  end
+
+  def setup_new_parser
+    self.processor = parser_class.new
+    self.lex = processor.lexer
+  end
+
+  def setup_lexer_class parser_class
+    self.parser_class = parser_class
+    setup_new_parser
+    setup_lexer "blah blah"
   end
 
   def test_advance
     assert @lex.advance # blah
     assert @lex.advance # blah
-    deny   @lex.advance # nada
+    refute @lex.advance # nada
+  end
+
+  def test_unicode_ident
+    s = "@\u1088\u1077\u1093\u1072"
+    util_lex_token(s.dup,
+                   :tIVAR, s.dup)
   end
 
   def test_read_escape
@@ -37,6 +63,11 @@ class TestRubyLexer < MiniTest::Unit::TestCase
     util_escape "\010", "b"
     util_escape " ",    "s"
     util_escape "q",    "q" # plain vanilla escape
+
+    util_escape "8", "8" # ugh... mri... WHY?!?
+    util_escape "9", "9" # ugh... mri... WHY?!?
+
+    util_escape "$",    "444" # ugh
   end
 
   def test_read_escape_c
@@ -125,6 +156,381 @@ class TestRubyLexer < MiniTest::Unit::TestCase
     util_lex_token "=>", :tASSOC, "=>"
   end
 
+  def test_yylex_label__18
+    setup_lexer_class Ruby18Parser
+
+    util_lex_token("{a:",
+                   :tLBRACE,     "{",
+                   :tIDENTIFIER, "a",
+                   :tSYMBEG,     ":")
+  end
+
+  def test_yylex_label_in_params__18
+    setup_lexer_class Ruby18Parser
+
+    util_lex_token("foo(a:",
+                   :tIDENTIFIER, "foo",
+                   :tLPAREN2,    "(",
+                   :tIDENTIFIER, "a",
+                   :tSYMBEG,     ":")
+  end
+
+  def test_yylex_label__19
+    setup_lexer_class Ruby19Parser
+
+    util_lex_token("{a:",
+                   :tLBRACE, "{",
+                   :tLABEL,  "a")
+  end
+
+  def test_yylex_label_in_params__19
+    setup_lexer_class Ruby19Parser
+
+    util_lex_token("foo(a:",
+                   :tIDENTIFIER, "foo",
+                   :tLPAREN2,    "(",
+                   :tLABEL,      "a")
+  end
+
+  def assert_next_lexeme token=nil, value=nil, state=nil, paren=nil, brace=nil
+    assert @lex.advance, "no more tokens"
+
+    msg = message {
+      act = [@lex.token, @lex.yacc_value, @lex.lex_state,
+             @lex.paren_nest, @lex.brace_nest]
+      exp = [token, value, state, paren, brace]
+      "#{exp.inspect} vs #{act.inspect}"
+    }
+
+    act_value = @lex.yacc_value
+    act_value = act_value.first if Array === act_value
+
+    assert_equal token, @lex.token,      msg
+    assert_equal value, act_value,       msg
+    assert_equal state, @lex.lex_state,  msg
+    assert_equal paren, @lex.paren_nest, msg if paren
+    assert_equal brace, @lex.brace_nest, msg if brace
+  end
+
+  def refute_lexeme
+    refute @lex.advance, "not empty: #{[@lex.token, @lex.yacc_value].inspect}"
+  end
+
+  def assert_lex input, exp_sexp, *args
+    setup_lexer input, exp_sexp
+
+    args.each_slice(5) do |token, value, state, paren, brace|
+      assert_next_lexeme token, value, state, paren, brace
+    end
+
+    refute_lexeme
+  end
+
+  def emulate_string_interpolation
+    lex_strterm = lexer.lex_strterm
+    string_nest = lexer.string_nest
+    brace_nest  = lexer.brace_nest
+
+    lexer.string_nest = 0
+    lexer.brace_nest  = 0
+    lexer.cond.push false
+    lexer.cmdarg.push false
+
+    lexer.lex_strterm = nil
+    lexer.lex_state = :expr_beg
+
+    yield
+
+    lexer.lex_state = :expr_endarg
+    assert_next_lexeme :tRCURLY,     "}",  :expr_endarg, 0
+
+    lexer.lex_strterm = lex_strterm
+    lexer.lex_state   = :expr_beg
+    lexer.string_nest = string_nest
+    lexer.brace_nest  = brace_nest
+
+    lexer.cond.lexpop
+    lexer.cmdarg.lexpop
+  end
+
+  def test_yylex_paren_string_parens_interpolated
+    setup_lexer('%((#{b}#{d}))',
+                s(:dstr,
+                  "(",
+                  s(:evstr, s(:call, nil, :b)),
+                  s(:evstr, s(:call, nil, :d)),
+                  s(:str, ")")))
+
+    assert_next_lexeme :tSTRING_BEG,     "%)", :expr_beg, 0, 0
+    assert_next_lexeme :tSTRING_CONTENT, "(",  :expr_beg, 0, 0
+    assert_next_lexeme :tSTRING_DBEG,    nil,  :expr_beg, 0, 0
+
+    emulate_string_interpolation do
+      assert_next_lexeme :tIDENTIFIER,   "b",  :expr_arg, 0, 0
+    end
+
+    assert_next_lexeme :tSTRING_DBEG,    nil,  :expr_beg, 0, 0
+
+    emulate_string_interpolation do
+      assert_next_lexeme :tIDENTIFIER,   "d",  :expr_arg, 0, 0
+    end
+
+    assert_next_lexeme :tSTRING_CONTENT, ")",  :expr_beg, 0, 0
+    assert_next_lexeme :tSTRING_END,     ")",  :expr_end, 0, 0
+
+    refute_lexeme
+  end
+
+  def test_yylex_paren_string_interpolated_regexp
+    setup_lexer('%( #{(/abcd/)} )',
+                s(:dstr, " ", s(:evstr, s(:lit, /abcd/)), s(:str, " ")))
+
+    assert_next_lexeme :tSTRING_BEG,       "%)",   :expr_beg, 0, 0
+    assert_next_lexeme :tSTRING_CONTENT,   " ",    :expr_beg, 0, 0
+    assert_next_lexeme :tSTRING_DBEG,      nil,    :expr_beg, 0, 0
+
+    emulate_string_interpolation do
+      assert_next_lexeme :tLPAREN,         "(",    :expr_beg, 1, 0
+      assert_next_lexeme :tREGEXP_BEG,     "/",    :expr_beg, 1, 0
+      assert_next_lexeme :tSTRING_CONTENT, "abcd", :expr_beg, 1, 0
+      assert_next_lexeme :tREGEXP_END,     "",     :expr_end, 1, 0
+      assert_next_lexeme :tRPAREN,         ")",    :expr_endfn, 0, 0
+    end
+
+    assert_next_lexeme :tSTRING_CONTENT,   " ",    :expr_beg, 0, 0
+    assert_next_lexeme :tSTRING_END,       ")",    :expr_end, 0, 0
+
+    refute_lexeme
+  end
+
+  def test_yylex_not_at_defn
+    assert_lex("def +@; end",
+               s(:defn, :+@, s(:args), s(:nil)),
+
+               :kDEF,   "def", :expr_fname, 0, 0,
+               :tUPLUS, "+@",  :expr_arg,   0, 0,
+               :tSEMI,  ";",   :expr_beg,   0, 0,
+               :kEND,   "end", :expr_end,   0, 0)
+
+    assert_lex("def !@; end",
+               s(:defn, :"!@", s(:args), s(:nil)),
+
+               :kDEF,   "def", :expr_fname, 0, 0,
+               :tUBANG, "!@",  :expr_arg,   0, 0,
+               :tSEMI,  ";",   :expr_beg,   0, 0,
+               :kEND,   "end", :expr_end,   0, 0)
+  end
+
+  def test_yylex_not_at_ivar
+    assert_lex("!@ivar",
+               s(:call, s(:ivar, :@ivar), :"!"),
+
+               :tBANG, "!",     :expr_beg, 0, 0,
+               :tIVAR, "@ivar", :expr_end, 0, 0)
+  end
+
+  def test_yylex_number_times_ident_times_return_number
+    assert_lex("1 * b * 3",
+               s(:call,
+                 s(:call, s(:lit, 1), :*, s(:call, nil, :b)),
+                 :*, s(:lit, 3)),
+
+               :tINTEGER,      1, :expr_end, 0, 0,
+               :tSTAR2,      "*", :expr_beg, 0, 0,
+               :tIDENTIFIER, "b", :expr_arg, 0, 0,
+               :tSTAR2,      "*", :expr_beg, 0, 0,
+               :tINTEGER,      3, :expr_end, 0, 0)
+
+    assert_lex("1 * b *\n 3",
+               s(:call,
+                 s(:call, s(:lit, 1), :*, s(:call, nil, :b)),
+                 :*, s(:lit, 3)),
+
+               :tINTEGER,      1, :expr_end, 0, 0,
+               :tSTAR2,      "*", :expr_beg, 0, 0,
+               :tIDENTIFIER, "b", :expr_arg, 0, 0,
+               :tSTAR2,      "*", :expr_beg, 0, 0,
+               :tINTEGER,      3, :expr_end, 0, 0)
+  end
+
+  def test_yylex_paren_string_parens_interpolated_regexp
+    setup_lexer('%((#{(/abcd/)}))',
+                s(:dstr, "(", s(:evstr, s(:lit, /abcd/)), s(:str, ")")))
+
+    assert_next_lexeme :tSTRING_BEG,       "%)",   :expr_beg, 0, 0
+    assert_next_lexeme :tSTRING_CONTENT,   "(",    :expr_beg, 0, 0
+
+    assert_next_lexeme :tSTRING_DBEG,       nil,   :expr_beg, 0, 0
+
+    emulate_string_interpolation do
+      assert_next_lexeme :tLPAREN,         "(",    :expr_beg, 1, 0
+      assert_next_lexeme :tREGEXP_BEG,     "/",    :expr_beg, 1, 0
+      assert_next_lexeme :tSTRING_CONTENT, "abcd", :expr_beg, 1, 0
+      assert_next_lexeme :tREGEXP_END,     "",     :expr_end, 1, 0
+      assert_next_lexeme :tRPAREN,         ")",    :expr_endfn, 0, 0
+    end
+
+    assert_next_lexeme :tSTRING_CONTENT,   ")",    :expr_beg, 0, 0
+    assert_next_lexeme :tSTRING_END,       ")",    :expr_end, 0, 0
+
+    refute_lexeme
+  end
+
+  def test_yylex_method_parens_chevron
+    assert_lex("a()<<1",
+               s(:call, s(:call, nil, :a), :<<, s(:lit, 1)),
+               :tIDENTIFIER, "a",   :expr_cmdarg, 0, 0,
+               :tLPAREN2,    "(",   :expr_beg,    1, 0,
+               :tRPAREN,     ")",   :expr_endfn,  0, 0,
+               :tLSHFT,      "<<" , :expr_beg,    0, 0,
+               :tINTEGER,    1,     :expr_end,    0, 0)
+  end
+
+  def test_yylex_lambda_args__20
+    setup_lexer_class Ruby20Parser
+
+    assert_lex("-> (a) { }",
+               s(:iter, s(:call, nil, :lambda),
+                 s(:args, :a)),
+
+               :tLAMBDA,     nil, :expr_endfn,  0, 0,
+               :tLPAREN2,    "(", :expr_beg,    1, 0,
+               :tIDENTIFIER, "a", :expr_arg,    1, 0,
+               :tRPAREN,     ")", :expr_endfn,  0, 0,
+               :tLCURLY,     "{", :expr_beg,    0, 1,
+               :tRCURLY,     "}", :expr_endarg, 0, 0)
+  end
+
+  def test_yylex_lambda_args_opt__20
+    setup_lexer_class Ruby20Parser
+
+    assert_lex("-> (a=nil) { }",
+               s(:iter, s(:call, nil, :lambda),
+                 s(:args, s(:lasgn, :a, s(:nil)))),
+
+               :tLAMBDA,     nil, :expr_endfn,  0, 0,
+               :tLPAREN2,    "(", :expr_beg,    1, 0,
+               :tIDENTIFIER, "a", :expr_arg,    1, 0,
+               :tEQL,        "=", :expr_beg,    1, 0,
+               :kNIL,        "nil", :expr_end,    1, 0,
+               :tRPAREN,     ")", :expr_endfn,  0, 0,
+               :tLCURLY,     "{", :expr_beg,    0, 1,
+               :tRCURLY,     "}", :expr_endarg, 0, 0)
+  end
+
+  def test_yylex_lambda_hash__20
+    setup_lexer_class Ruby20Parser
+
+    assert_lex("-> (a={}) { }",
+               s(:iter, s(:call, nil, :lambda),
+                 s(:args, s(:lasgn, :a, s(:hash)))),
+
+               :tLAMBDA,     nil, :expr_endfn,  0, 0,
+               :tLPAREN2,    "(", :expr_beg,    1, 0,
+               :tIDENTIFIER, "a", :expr_arg,    1, 0,
+               :tEQL,        "=", :expr_beg,    1, 0,
+               :tLBRACE,     "{", :expr_beg,    1, 1,
+               :tRCURLY,     "}", :expr_endarg, 1, 0,
+               :tRPAREN,     ")", :expr_endfn,  0, 0,
+               :tLCURLY,     "{", :expr_beg,    0, 1,
+               :tRCURLY,     "}", :expr_endarg, 0, 0)
+  end
+
+  def test_yylex_iter_array_curly
+    assert_lex("f :a, [:b] { |c, d| }", # yes, this is bad code
+               s(:iter,
+                 s(:call, nil, :f, s(:lit, :a), s(:array, s(:lit, :b))),
+                 s(:args, :c, :d)),
+
+               :tIDENTIFIER, "f", :expr_cmdarg, 0, 0,
+               :tSYMBOL,     "a", :expr_end,    0, 0,
+               :tCOMMA,      ",", :expr_beg,    0, 0,
+               :tLBRACK,     "[", :expr_beg,    1, 0,
+               :tSYMBOL,     "b", :expr_end,    1, 0,
+               :tRBRACK,     "]", :expr_endarg, 0, 0,
+               :tLBRACE_ARG, "{", :expr_beg,    0, 1,
+               :tPIPE,       "|", :expr_beg,    0, 1,
+               :tIDENTIFIER, "c", :expr_arg,    0, 1,
+               :tCOMMA,      ",", :expr_beg,    0, 1,
+               :tIDENTIFIER, "d", :expr_arg,    0, 1,
+               :tPIPE,       "|", :expr_beg,    0, 1,
+               :tRCURLY,     "}", :expr_endarg, 0, 0)
+  end
+
+  def test_yylex_const_call_same_name
+    assert_lex("X = a { }; b { f :c }",
+               s(:block,
+                 s(:cdecl, :X, s(:iter, s(:call, nil, :a), s(:args))),
+                 s(:iter,
+                   s(:call, nil, :b),
+                   s(:args),
+                   s(:call, nil, :f, s(:lit, :c)))),
+
+               :tCONSTANT,   "X", :expr_cmdarg, 0, 0,
+               :tEQL,        "=", :expr_beg,    0, 0,
+               :tIDENTIFIER, "a", :expr_arg,    0, 0,
+               :tLCURLY,     "{", :expr_beg,    0, 1,
+               :tRCURLY,     "}", :expr_endarg, 0, 0,
+               :tSEMI,       ";", :expr_beg,    0, 0,
+
+               :tIDENTIFIER, "b", :expr_cmdarg, 0, 0,
+               :tLCURLY,     "{", :expr_beg,    0, 1,
+               :tIDENTIFIER, "f", :expr_cmdarg, 0, 1, # different
+               :tSYMBOL,     "c", :expr_end,    0, 1,
+               :tRCURLY,     "}", :expr_endarg, 0, 0)
+
+    assert_lex("X = a { }; b { X :c }",
+               s(:block,
+                 s(:cdecl, :X, s(:iter, s(:call, nil, :a), s(:args))),
+                 s(:iter,
+                   s(:call, nil, :b),
+                   s(:args),
+                   s(:call, nil, :X, s(:lit, :c)))),
+
+               :tCONSTANT,   "X", :expr_cmdarg, 0, 0,
+               :tEQL,        "=", :expr_beg,    0, 0,
+               :tIDENTIFIER, "a", :expr_arg,    0, 0,
+               :tLCURLY,     "{", :expr_beg,    0, 1,
+               :tRCURLY,     "}", :expr_endarg, 0, 0,
+               :tSEMI,       ";", :expr_beg,    0, 0,
+
+               :tIDENTIFIER, "b", :expr_cmdarg, 0, 0,
+               :tLCURLY,     "{", :expr_beg,    0, 1,
+               :tCONSTANT,   "X", :expr_cmdarg, 0, 1, # same
+               :tSYMBOL,     "c", :expr_end,    0, 1,
+               :tRCURLY,     "}", :expr_endarg, 0, 0)
+  end
+
+  def test_yylex_lasgn_call_same_name
+    assert_lex("a = b.c :d => 1",
+               s(:lasgn, :a,
+                 s(:call, s(:call, nil, :b), :c,
+                   s(:hash, s(:lit, :d), s(:lit, 1)))),
+
+               :tIDENTIFIER, "a", :expr_cmdarg, 0, 0,
+               :tEQL,        "=", :expr_beg,    0, 0,
+               :tIDENTIFIER, "b", :expr_arg,    0, 0,
+               :tDOT,        ".", :expr_dot,    0, 0,
+               :tIDENTIFIER, "c", :expr_arg,    0, 0, # different
+               :tSYMBOL,     "d", :expr_end,    0, 0,
+               :tASSOC,      "=>", :expr_beg,   0, 0,
+               :tINTEGER,      1, :expr_end,    0, 0)
+
+    assert_lex("a = b.a :d => 1",
+               s(:lasgn, :a,
+                 s(:call, s(:call, nil, :b), :a,
+                   s(:hash, s(:lit, :d), s(:lit, 1)))),
+
+               :tIDENTIFIER, "a", :expr_cmdarg, 0, 0,
+               :tEQL,        "=", :expr_beg,    0, 0,
+               :tIDENTIFIER, "b", :expr_arg,    0, 0,
+               :tDOT,        ".", :expr_dot,    0, 0,
+               :tIDENTIFIER, "a", :expr_arg,    0, 0, # same as lvar
+               :tSYMBOL,     "d", :expr_end,    0, 0,
+               :tASSOC,      "=>", :expr_beg,   0, 0,
+               :tINTEGER,      1, :expr_end,    0, 0)
+  end
+
   def test_yylex_back_ref
     util_lex_token("[$&, $`, $', $+]",
                    :tLBRACK,   "[",
@@ -280,7 +686,7 @@ class TestRubyLexer < MiniTest::Unit::TestCase
   end
 
   def test_yylex_cvar_bad
-    assert_raises SyntaxError do
+    assert_raises RubyParser::SyntaxError do
       util_lex_token "@@1"
     end
   end
@@ -589,6 +995,7 @@ class TestRubyLexer < MiniTest::Unit::TestCase
   end
 
   def test_yylex_heredoc_double_interp
+    # TODO: convert to assert_lex
     util_lex_token("a = <<\"EOF\"\n#x a \#@a b \#$b c \#{3} \nEOF\n",
                    :tIDENTIFIER,     "a",
                    :tEQL,              "=",
@@ -604,13 +1011,21 @@ class TestRubyLexer < MiniTest::Unit::TestCase
                    :tNL,             nil)
   end
 
+  def test_yylex_heredoc_empty
+    util_lex_token("<<\"\"\n\#{x}\nblah2\n\n",
+                   :tSTRING_BEG,     "\"",
+                   :tSTRING_DBEG,    "\#{",
+                   :tSTRING_CONTENT, "x}\nblah2\n",
+                   :tSTRING_END,     "",
+                   :tNL,             nil)
+  end
+
   def test_yylex_heredoc_none
     util_lex_token("a = <<EOF\nblah\nblah\nEOF",
                    :tIDENTIFIER,     "a",
                    :tEQL,              "=",
                    :tSTRING_BEG,     "\"",
                    :tSTRING_CONTENT, "blah\nblah\n",
-                   :tSTRING_CONTENT, "",
                    :tSTRING_END,     "EOF",
                    :tNL,             nil)
   end
@@ -628,7 +1043,6 @@ class TestRubyLexer < MiniTest::Unit::TestCase
                    :tEQL,              "=",
                    :tSTRING_BEG,     "\"",
                    :tSTRING_CONTENT, "blah\nblah\n",
-                   :tSTRING_CONTENT, "",
                    :tSTRING_END,     "EOF",
                    :tNL,             nil)
   end
@@ -693,10 +1107,18 @@ class TestRubyLexer < MiniTest::Unit::TestCase
     util_lex_fname "<=>", :tCMP
   end
 
-  def test_yylex_identifier_def
+  def test_yylex_identifier_def__18
+    setup_lexer_class Ruby18Parser
+
     util_lex_fname "identifier", :tIDENTIFIER, :expr_end
   end
 
+  def test_yylex_identifier_def__1920
+    setup_lexer_class Ruby19Parser
+
+    util_lex_fname "identifier", :tIDENTIFIER, :expr_endfn
+  end
+
   def test_yylex_identifier_eh
     util_lex_token("identifier?", :tFID, "identifier?")
   end
@@ -708,14 +1130,38 @@ class TestRubyLexer < MiniTest::Unit::TestCase
                    :tASSOC, "=>")
   end
 
+  def test_yylex_identifier_equals3
+    # @lex.lex_state = :expr_fname
+    util_lex_token(":a===b",
+                   :tSYMBOL, "a",
+                   :tEQQ, "===",
+                   :tIDENTIFIER, "b")
+  end
+
+  def test_yylex_identifier_equals_equals_arrow
+    # @lex.lex_state = :expr_fname
+    util_lex_token(":a==>b",
+                   :tSYMBOL, "a=",
+                   :tASSOC, "=>",
+                   :tIDENTIFIER, "b")
+  end
+
   def test_yylex_identifier_equals_caret
     util_lex_fname "^", :tCARET
   end
 
-  def test_yylex_identifier_equals_def
+  def test_yylex_identifier_equals_def__18
+    setup_lexer_class Ruby18Parser
+
     util_lex_fname "identifier=", :tIDENTIFIER, :expr_end
   end
 
+  def test_yylex_identifier_equals_def__1920
+    setup_lexer_class Ruby19Parser
+
+    util_lex_fname "identifier=", :tIDENTIFIER, :expr_endfn
+  end
+
   def test_yylex_identifier_equals_def2
     util_lex_fname "==", :tEQ
   end
@@ -805,14 +1251,30 @@ class TestRubyLexer < MiniTest::Unit::TestCase
     util_bad_token "0d42__24"
   end
 
-  def test_yylex_integer_eh_a
-    util_lex_token '?a', :tINTEGER, 97
+  def test_yylex_question_eh_a__18
+    setup_lexer_class Ruby18Parser
+
+    util_lex_token "?a", :tINTEGER, 97
   end
 
-  def test_yylex_integer_eh_escape_M_escape_C
+  def test_yylex_question_eh_a__19
+    setup_lexer_class Ruby19Parser
+
+    util_lex_token '?a', :tSTRING, "a"
+  end
+
+  def test_yylex_question_eh_escape_M_escape_C__18
+    setup_lexer_class Ruby18Parser
+
     util_lex_token '?\M-\C-a', :tINTEGER, 129
   end
 
+  def test_yylex_question_eh_escape_M_escape_C__19
+    setup_lexer_class Ruby19Parser
+
+    util_lex_token '?\M-\C-a', :tSTRING, "\M-\C-a"
+  end
+
   def test_yylex_integer_hex
     util_lex_token "0x2a", :tINTEGER, 42
   end
@@ -833,6 +1295,10 @@ class TestRubyLexer < MiniTest::Unit::TestCase
     util_bad_token "08"
   end
 
+  def test_yylex_integer_oct_bad_range2
+    util_bad_token "08"
+  end
+
   def test_yylex_integer_oct_bad_underscores
     util_bad_token "01__23"
   end
@@ -896,6 +1362,10 @@ class TestRubyLexer < MiniTest::Unit::TestCase
     util_bad_token "@1"
   end
 
+  def test_yylex_ivar_bad_0_length
+    util_bad_token "1+@\n", :tINTEGER, 1, :tPLUS, "+"
+  end
+
   def test_yylex_keyword_expr
     @lex.lex_state = :expr_endarg
 
@@ -978,11 +1448,20 @@ class TestRubyLexer < MiniTest::Unit::TestCase
     util_lex_token(" (", :tLPAREN_ARG, "(")
   end
 
-  def test_yylex_open_bracket_exprarg
+  def test_yylex_open_bracket_exprarg__18
+    setup_lexer_class Ruby18Parser
+
     @lex.lex_state = :expr_arg
     util_lex_token(" (", :tLPAREN2, "(")
   end
 
+  def test_yylex_open_bracket_exprarg__19
+    setup_lexer_class Ruby19Parser
+
+    @lex.lex_state = :expr_arg
+    util_lex_token(" (", :tLPAREN_ARG, "(")
+  end
+
   def test_yylex_open_curly_bracket
     util_lex_token("{",
                    :tLBRACE, "{")
@@ -1028,7 +1507,7 @@ class TestRubyLexer < MiniTest::Unit::TestCase
   def test_yylex_open_square_bracket_meth
     util_lex_token("m[3]",
                    :tIDENTIFIER, "m",
-                   "[", "[",
+                   :tLBRACK2, "[",
                    :tINTEGER, 3,
                    :tRBRACK, "]")
   end
@@ -1084,15 +1563,69 @@ class TestRubyLexer < MiniTest::Unit::TestCase
     util_lex_token "+@", :tUPLUS, "+@"
   end
 
+  def test_yylex_not_unary_method
+    skip "not yet"
+
+    @lex.lex_state = :expr_fname
+    util_lex_token "!@", :tUBANG, "!@"
+  end
+
+  def test_yylex_numbers
+    util_lex_token "0b10", :tINTEGER, 2
+    util_lex_token "0B10", :tINTEGER, 2
+
+    util_lex_token "0d10", :tINTEGER, 10
+    util_lex_token "0D10", :tINTEGER, 10
+
+    util_lex_token "0x10", :tINTEGER, 16
+    util_lex_token "0X10", :tINTEGER, 16
+
+    util_lex_token "0o10", :tINTEGER, 8
+    util_lex_token "0O10", :tINTEGER, 8
+    util_lex_token "0o",   :tINTEGER, 0
+    util_lex_token "0O",   :tINTEGER, 0
+
+    util_lex_token "0o",   :tINTEGER, 0
+    util_lex_token "0O",   :tINTEGER, 0
+
+    util_lex_token "0",    :tINTEGER, 0
+
+    util_bad_token "0x"
+    util_bad_token "0X"
+    util_bad_token "0b"
+    util_bad_token "0B"
+    util_bad_token "0d"
+    util_bad_token "0D"
+
+    util_bad_token "08"
+    util_bad_token "09"
+    util_bad_token "0o8"
+    util_bad_token "0o9"
+    util_bad_token "0O8"
+    util_bad_token "0O9"
+
+    util_bad_token "1_e1"
+    util_bad_token "1_.1"
+    util_bad_token "1__1"
+  end
+
   def test_yylex_plus_unary_number
     util_lex_token("+42",
                    :tINTEGER, 42)
   end
 
-  def test_yylex_question
+  def test_yylex_question__18
+    setup_lexer_class Ruby18Parser
+
     util_lex_token "?*", :tINTEGER, 42
   end
 
+  def test_yylex_question__19
+    setup_lexer_class Ruby19Parser
+
+    util_lex_token "?*", :tSTRING, "*"
+  end
+
   def test_yylex_question_bad_eos
     util_bad_token "?"
   end
@@ -1106,7 +1639,9 @@ class TestRubyLexer < MiniTest::Unit::TestCase
     util_lex_token "?\f", :tEH, "?"
   end
 
-  def test_yylex_question_ws_backslashed
+  def test_yylex_question_ws_backslashed__18
+    setup_lexer_class Ruby18Parser
+
     @lex.lex_state = :expr_beg
     util_lex_token "?\\ ", :tINTEGER, 32
     @lex.lex_state = :expr_beg
@@ -1121,6 +1656,23 @@ class TestRubyLexer < MiniTest::Unit::TestCase
     util_lex_token "?\\f", :tINTEGER, 12
   end
 
+  def test_yylex_question_ws_backslashed__19
+    setup_lexer_class Ruby19Parser
+
+    @lex.lex_state = :expr_beg
+    util_lex_token "?\\ ", :tSTRING, " "
+    @lex.lex_state = :expr_beg
+    util_lex_token "?\\n", :tSTRING, "\n"
+    @lex.lex_state = :expr_beg
+    util_lex_token "?\\t", :tSTRING, "\t"
+    @lex.lex_state = :expr_beg
+    util_lex_token "?\\v", :tSTRING, "\v"
+    @lex.lex_state = :expr_beg
+    util_lex_token "?\\r", :tSTRING, "\r"
+    @lex.lex_state = :expr_beg
+    util_lex_token "?\\f", :tSTRING, "\f"
+  end
+
   def test_yylex_rbracket
     util_lex_token "]", :tRBRACK, "]"
   end
@@ -1469,8 +2021,17 @@ class TestRubyLexer < MiniTest::Unit::TestCase
   end
 
   def test_yylex_string_double_escape_M
+    chr = "\341"
+    chr.force_encoding("UTF-8") if RubyLexer::RUBY19
+
     util_lex_token('"\\M-a"',
-                   :tSTRING, "\341")
+                   :tSTRING, chr)
+  end
+
+  def test_why_does_ruby_hate_me?
+    util_lex_token('"Nl%\000\000A\000\999"', # you should be ashamed
+                   :tSTRING,
+                   ["Nl%", "\x00", "\x00", "A", "\x00", "999"].join)
   end
 
   def test_yylex_string_double_escape_M_backslash
@@ -1536,6 +2097,11 @@ class TestRubyLexer < MiniTest::Unit::TestCase
                    :tSTRING, "n = ABC")
   end
 
+  def test_yylex_string_double_escape_octal_fucked
+    util_lex_token('"n = \\444"',
+                   :tSTRING, "n = $")
+  end
+
   def test_yylex_string_double_interp
     util_lex_token("\"blah #x a \#@a b \#$b c \#{3} # \"",
                    :tSTRING_BEG,     "\"",
@@ -1569,6 +2135,30 @@ class TestRubyLexer < MiniTest::Unit::TestCase
                    :tSTRING, "\000")
   end
 
+  def test_yylex_string_pct_i
+    util_lex_token("%i[s1 s2\ns3]",
+                   :tQSYMBOLS_BEG,   "%i[",
+                   :tSTRING_CONTENT, "s1",
+                   :tSPACE,              nil,
+                   :tSTRING_CONTENT, "s2",
+                   :tSPACE,              nil,
+                   :tSTRING_CONTENT, "s3",
+                   :tSPACE,              nil,
+                   :tSTRING_END,     nil)
+  end
+
+  def test_yylex_string_pct_I
+    util_lex_token("%I[s1 s2\ns3]",
+                   :tSYMBOLS_BEG,    "%I[",
+                   :tSTRING_CONTENT, "s1",
+                   :tSPACE,              nil,
+                   :tSTRING_CONTENT, "s2",
+                   :tSPACE,              nil,
+                   :tSTRING_CONTENT, "s3",
+                   :tSPACE,              nil,
+                   :tSTRING_END,     nil)
+  end
+
   def test_yylex_string_pct_Q
     util_lex_token("%Q[s1 s2]",
                    :tSTRING_BEG,     "%Q[",
@@ -1614,7 +2204,7 @@ class TestRubyLexer < MiniTest::Unit::TestCase
 
   def test_yylex_string_pct_w
     util_bad_token("%w[s1 s2 ",
-                   :tAWORDS_BEG,     "%w[",
+                   :tQWORDS_BEG,     "%w[",
                    :tSTRING_CONTENT, "s1",
                    :tSPACE,              nil,
                    :tSTRING_CONTENT, "s2",
@@ -1623,7 +2213,7 @@ class TestRubyLexer < MiniTest::Unit::TestCase
 
   def test_yylex_string_pct_w_bs_nl
     util_lex_token("%w[s1 \\\ns2]",
-                   :tAWORDS_BEG,     "%w[",
+                   :tQWORDS_BEG,     "%w[",
                    :tSTRING_CONTENT, "s1",
                    :tSPACE,              nil,
                    :tSTRING_CONTENT, "\ns2",
@@ -1633,7 +2223,7 @@ class TestRubyLexer < MiniTest::Unit::TestCase
 
   def test_yylex_string_pct_w_bs_sp
     util_lex_token("%w[s\\ 1 s\\ 2]",
-                   :tAWORDS_BEG,     "%w[",
+                   :tQWORDS_BEG,     "%w[",
                    :tSTRING_CONTENT, "s 1",
                    :tSPACE,              nil,
                    :tSTRING_CONTENT, "s 2",
@@ -1641,14 +2231,6 @@ class TestRubyLexer < MiniTest::Unit::TestCase
                    :tSTRING_END,     nil)
   end
 
-  def test_yylex_string_pct_w_tab
-    util_lex_token("%w[abc\tdef]",
-                   :tAWORDS_BEG,      "%w[",
-                   :tSTRING_CONTENT, "abc\tdef",
-                   :tSPACE,              nil,
-                   :tSTRING_END,     nil)
-  end
-
   def test_yylex_string_single
     util_lex_token("'string'",
                    :tSTRING, "string")
@@ -1729,7 +2311,7 @@ class TestRubyLexer < MiniTest::Unit::TestCase
 
   def test_yylex_underscore_end
     @lex.src = "__END__\n"
-    deny @lex.advance
+    refute @lex.advance
   end
 
   def test_yylex_uplus
@@ -1785,22 +2367,40 @@ class TestRubyLexer < MiniTest::Unit::TestCase
                    :tFLOAT, 0.0)
   end
 
+  def test_pct_w_backslashes
+    ["\t", "\n", "\r", "\v", "\f"].each do |char|
+      next if !RubyLexer::RUBY19 and char == "\v"
+
+      assert_lex("%w[foo#{char}bar]",
+                 s(:array, s(:str, "foo"), s(:str, "bar")),
+
+                 :tQWORDS_BEG,     "%w[", :expr_beg, 0, 0,
+                 :tSTRING_CONTENT, "foo", :expr_beg, 0, 0,
+                 :tSPACE,          nil,   :expr_beg, 0, 0,
+                 :tSTRING_CONTENT, "bar", :expr_beg, 0, 0,
+                 :tSPACE,          nil,   :expr_beg, 0, 0,
+                 :tSTRING_END,     nil,   :expr_end, 0, 0)
+    end
+
+    # flunk "Not yet"
+  end
+
   ############################################################
 
   def util_bad_token s, *args
-    assert_raises SyntaxError do
+    assert_raises RubyParser::SyntaxError do
       util_lex_token s, *args
     end
   end
 
   def util_escape expected, input
     @lex.src = input
-    assert_equal expected, @lex.read_escape
+    assert_equal expected, @lex.read_escape, input
   end
 
   def util_escape_bad input
     @lex.src = input
-    assert_raises SyntaxError do
+    assert_raises RubyParser::SyntaxError do
       @lex.read_escape
     end
   end
@@ -1808,7 +2408,11 @@ class TestRubyLexer < MiniTest::Unit::TestCase
   def util_lex_fname name, type, end_state = :expr_arg
     @lex.lex_state = :expr_fname # can only set via parser's defs
 
-    util_lex_token("def #{name} ", :kDEF, "def", type, name)
+    assert_lex("def #{name} ",
+               nil,
+
+               :kDEF, "def", :expr_fname, 0, 0,
+               type, name, end_state, 0, 0)
 
     assert_equal end_state, @lex.lex_state
   end
@@ -1820,10 +2424,10 @@ class TestRubyLexer < MiniTest::Unit::TestCase
       token = args.shift
       value = args.shift
       assert @lex.advance, "no more tokens"
-      assert_equal [token, value], [@lex.token, [@lex.yacc_value].flatten.first]
+      # assert_equal [token, value].map(&:encoding), [@lex.token, [@lex.yacc_value].flatten.first].map(&:encoding), input # TODO
+      assert_equal [token, value], [@lex.token, [@lex.yacc_value].flatten.first], input
     end
 
-    deny @lex.advance, "must be empty, but had #{[@lex.token, @lex.yacc_value].inspect}"
+    refute @lex.advance, "must be empty, but had #{[@lex.token, @lex.yacc_value].inspect}"
   end
 end
-
